/***************************************************************************
 *   Copyright (C) 2004 by Raphael Langerhorst                             *
 *   raphael-langerhorst@gmx.at                                            *
 *                                                                         *
 *   Permission is hereby granted, free of charge, to any person obtaining *
 *   a copy of this software and associated documentation files (the       *
 *   "Software"), to deal in the Software without restriction, including   *
 *   without limitation the rights to use, copy, modify, merge, publish,   *
 *   distribute, sublicense, and/or sell copies of the Software, and to    *
 *   permit persons to whom the Software is furnished to do so, subject to *
 *   the following conditions:                                             *
 *                                                                         *
 *   The above copyright notice and this permission notice shall be        *
 *   included in all copies or substantial portions of the Software.       *
 *                                                                         *
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,       *
 *   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF    *
 *   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.*
 *   IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR     *
 *   OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, *
 *   ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR *
 *   OTHER DEALINGS IN THE SOFTWARE.                                       *
 ***************************************************************************/

#ifndef GBEGACTIVEELEMENTINFLUENCE_H
#define GBEGACTIVEELEMENTINFLUENCE_H

#include "GPositionalInfluence.h"

namespace GBE {

/**
  \class GActiveElementInfluence GActiveElementInfluence.h
  \brief Active element influences causes all elements in the affected area to be in-memory and executed.
  @author Raphael Langerhorst

  Active elements can either be completely intelligent agent driven
  or controlled by users. Every element that plays an "active" role
  in the universe, like users of the system, should radiate this
  kind of influence because it is used by the dynamic world generation
  framework to determine which parts (or elements) of the universe
  have to be kept in memory and executed.
  
  This influence inherits GPositionalInfluence in order to allow easy
  calculation of the area center since position and parent (element context)
  are known.
*/

class GActiveElementInfluence : public GBE::GPositionalInfluence
{
  protected:
    /**
     * Determines the area size around the influence source
     * which has to be kept active by the system.
     */
    double SenseRadius;
    
    /**
     * Determines the area size around the influence source
     * which has to be kept active by the system even if the
     * receiving element is one the same hierarchy level,
     * e.g. looking into a house from outside or onto the
     * planet surface from space (=without having entered
     * the planet or house).
     */
    double SenseRadiusDetail;
    
  public:
  
    /**
     * Constructor.
     */
    GActiveElementInfluence(const GCS::GElementID& source, const GCS::GEnergy& energy, const GCS::GVector3& position, const GCS::GVector3& rotation, double radius, const GCS::GElementID& parent, double sense_radius, double sense_radius_detail);

    /**
     * Destructor.
     */
    virtual ~GActiveElementInfluence();
    
    /**
     * @returns the sense radius of the influence source.
     */
    double senseRadius() const;
    
    double senseRadiusDetail() const;

};

};

#endif
