/***************************************************************************
 *   Copyright (C) 2004 by Raphael Langerhorst                             *
 *   raphael-langerhorst@gmx.at                                            *
 *                                                                         *
 *   Permission is hereby granted, free of charge, to any person obtaining *
 *   a copy of this software and associated documentation files (the       *
 *   "Software"), to deal in the Software without restriction, including   *
 *   without limitation the rights to use, copy, modify, merge, publish,   *
 *   distribute, sublicense, and/or sell copies of the Software, and to    *
 *   permit persons to whom the Software is furnished to do so, subject to *
 *   the following conditions:                                             *
 *                                                                         *
 *   The above copyright notice and this permission notice shall be        *
 *   included in all copies or substantial portions of the Software.       *
 *                                                                         *
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,       *
 *   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF    *
 *   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.*
 *   IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR     *
 *   OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, *
 *   ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR *
 *   OTHER DEALINGS IN THE SOFTWARE.                                       *
 ***************************************************************************/

#ifndef GBEGINVESTIGATIONAGENTH
#define GBEGINVESTIGATIONAGENTH

#include <GAgent.h>
#include <GElementID.h>

#include <GElementInfluence.h>

#include <GWEInterface.h>

#include <qtimer.h>
#include <qvaluelist.h>
#include <qstringlist.h>

namespace GBE {

/**
  \class GInvestigationAgent GInvestigationAgent.h
  \brief Allows for environment investigation and analysation.
  @author Raphael Langerhorst
  
  This agent allows for easy environment analysation.
  It periodically emits information about itself as well
  as received influence.
  Not all information is emitted from signals of this class,
  the GAgent base class already offers some signals like
  energy change or form change signals, these are also used!
  
  The investigation agent uses a direct pointer to a GWE to
  access data from other elements, this is NOT standard and
  shouldn't be done in other agents.
*/
class GInvestigationAgent : public GCS::GAgent
{
  Q_OBJECT
  public:
    GInvestigationAgent(GWE::GWEInterface* gwe, int update_interval_msec = 500, QObject* parent = 0, const char* name = 0);

    ~GInvestigationAgent();

  protected:
  
    /**
     * Update timer used for energyChanged and formChanged signals.
     */
    QTimer UpdateTimer;
  
    /**
     * All elements that get in contact with this investigation element
     * (i.e. from which influences where received) are listed here.
     * @todo: older entries should be removed(?)
     */
    QValueList<GCS::GElementID> KnownElements;
    
    /**
     * If set to true and the changed signals are emitted, then
     * the known elements are also emited with the
     * knownElementsListChanged() signal.
     * @see UpdateTimer, KnownElements, knownElementsListChanged().
     */
    bool KnownElementsEmitPending;
  
    /**
     * Used to get exact locations of elements (for movement,...).
     */
    GWE::GWEInterface* Gwe;
    
    /**
     * Determines the time between emitting of updates.
     */
    int UpdateInterval;
    
    /**
     * The investigation element is following this target
     */
    unsigned long CurrentTarget;
    
    /**
     * Sets the current speed of the element, the agent will try force
     * this onto the form's speed attribute.
     * @note CurrentTarget must be set to 0 for this to work, otherwise
     *       the element will follow its current target.
     */
    GCS::GVector3 CurrentSpeed;
  
    /**
     * Constantly accelerates towards the element with the ID of currentTarget.
     */
    virtual void run();

    /**
     * Here the agent basically adds the received influence amount
     * to EnergyAmountReceived.
     * @todo: this has to be improved as we want a usable sensor
     * @todo: that means an energylevel spectrum analysation would be good for example
     */
    virtual void receiveInfluence(GCS::GElementInfluence& influence);
    
  protected slots:
  
    /**
     * Sends updates about form and energy.
     * @note If you think that the signal of GElement can be used as well for
     * these changes, then you are correct in principal, but the agent is
     * simplier to use when all signals can be connected from one object.
     */
    virtual void sendChangedSignals();
    
  public slots:
  
    /**
     * Creates a standard GElementInfluence object and sends it to given ID,
     * if ID == 0 (default value) then it is radiated
     */
    virtual void createAndSendBasicInfluence(double fraction_of_own_energy, unsigned long targetID = 0);
    
    /**
     * Makes this element move to given ID and follow it
     * @see run(), CurrentTarget
     */
    virtual void followTarget(unsigned long destinationElementID);
    
    /**
     * Sets the speed attribute.
     * @see CurrentSpeed
     */
    virtual void setSpeed(const GCS::GVector3& new_speed);

  signals:
      
    /**
     * Emitted every time an influence had been received
     */
    void influenceReceived(const GCS::GElementInfluence& i);

    /**
     * Emitted every time the KnownElements list has changed
     */
    void knownElementsListChanged(const QStringList& known_elements);
    
    //note: there are standard signals already defined in GAgent,
    //      these are also emitted constantly to allow analysation
    //      especially energyChanged() and formChanged() are in use.
};

};

#endif
