/***************************************************************************
 *   Copyright (C) 2004 by Raphael Langerhorst                             *
 *   raphael-langerhorst@gmx.at                                            *
 *                                                                         *
 *   Permission is hereby granted, free of charge, to any person obtaining *
 *   a copy of this software and associated documentation files (the       *
 *   "Software"), to deal in the Software without restriction, including   *
 *   without limitation the rights to use, copy, modify, merge, publish,   *
 *   distribute, sublicense, and/or sell copies of the Software, and to    *
 *   permit persons to whom the Software is furnished to do so, subject to *
 *   the following conditions:                                             *
 *                                                                         *
 *   The above copyright notice and this permission notice shall be        *
 *   included in all copies or substantial portions of the Software.       *
 *                                                                         *
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,       *
 *   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF    *
 *   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.*
 *   IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR     *
 *   OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, *
 *   ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR *
 *   OTHER DEALINGS IN THE SOFTWARE.                                       *
 ***************************************************************************/

#include "GMoveAgent.h"

#include <GObject.h>
#include <GForm.h>
#include <GVector3.h>

#include <qmutex.h>

using namespace GCS;

namespace GBE {

GMoveAgent::GMoveAgent(double fraction_slow_down_per_sec, unsigned long update_interval_msec, QObject* parent, const char* name)
: GAgent(parent, name),
  SlowDownFraction(fraction_slow_down_per_sec),
  Interval(update_interval_msec)
{
  if (Interval <= 0)
    Interval = 1;
}


GMoveAgent::~GMoveAgent()
{
}

void GMoveAgent::setInterval(unsigned long new_interval_msec)
{
  if (new_interval_msec > 0)
    this->Interval = new_interval_msec;
}

void GMoveAgent::run()
{
  if (requestObject()->hasForm())
  {
    Time.start();
    while(!shutdown)
    {
      updateForm();
      msleep(Interval);
    }
    //initialize timer
    Time.start();
    setInterval(Interval);
  }
  else
  {
    qDebug("element %lu has no GForm object, GMoveAgent is useless in this case", getElementID().getID());
  }
}

void GMoveAgent::updateForm()
{
  double delta_t = Time.restart()*0.001;
  GForm* f = requestForm();
  
  QMutexLocker m(f);
 
  //slows down a bit 
  f->Dynamics.Speed.mul(1-this->SlowDownFraction*delta_t);
  f->Dynamics.RotationSpeed.mul(1-this->SlowDownFraction*delta_t);
  
  f->Position += f->Dynamics.Speed * delta_t;
  
  f->Rotation += f->Dynamics.RotationSpeed * delta_t;
  
  emit this->formChanged(*f);
}

void GMoveAgent::addSpeedImpulse(const GCS::GVector3& add_speed)
{
  if (requestObject()->hasForm())
  {
    GForm* f = requestForm();
    QMutexLocker m(f);
    
    //add speed to form:
    f->Dynamics.Speed.add(add_speed);
    
    //@todo: emit form changed signal? Do we handle dynamic data as form attributes? Coordinate this with element wide property storage
    //emit this->formChanged(*f);
  }
}

void GMoveAgent::addRotationImpulse(const GCS::GVector3& rotation_impulse)
{
  if (requestObject()->hasForm())
  {
    GForm* f = requestForm();
    QMutexLocker m(f);
    
    f->Dynamics.RotationSpeed.add(rotation_impulse);
  }
}

};
