/***************************************************************************
 *   Copyright (C) 2004 by Raphael Langerhorst                             *
 *   raphael-langerhorst@gmx.at                                            *
 *                                                                         *
 *   Permission is hereby granted, free of charge, to any person obtaining *
 *   a copy of this software and associated documentation files (the       *
 *   "Software"), to deal in the Software without restriction, including   *
 *   without limitation the rights to use, copy, modify, merge, publish,   *
 *   distribute, sublicense, and/or sell copies of the Software, and to    *
 *   permit persons to whom the Software is furnished to do so, subject to *
 *   the following conditions:                                             *
 *                                                                         *
 *   The above copyright notice and this permission notice shall be        *
 *   included in all copies or substantial portions of the Software.       *
 *                                                                         *
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,       *
 *   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF    *
 *   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.*
 *   IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR     *
 *   OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, *
 *   ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR *
 *   OTHER DEALINGS IN THE SOFTWARE.                                       *
 ***************************************************************************/

#include "GRadiatingAgent.h"
#include "GPositionalInfluence.h"
#include "GActiveElementInfluence.h"

#include <GObject.h>
#include <GElementInfluence.h>
#include <GEnergy.h>

using namespace GCS;

namespace GBE {

GRadiatingAgent::GRadiatingAgent(double fraction, double min_energy_amount, double max_energy_amount, unsigned long msleep, QObject* parent, const char* name)
: GAgent(parent, name),
  Fraction(fraction),
  MinEnergyAmount(min_energy_amount),
  MaxEnergyAmount(max_energy_amount),
  Msleep(msleep),
  ActiveElement(false),
  SenseDistance(0),
  SenseDistanceDetail(0)
{
  if (min_energy_amount > max_energy_amount)
  {
    qWarning("min_energy_amount is bigger than max_energy_amount!!!");
  }
  else if (min_energy_amount == max_energy_amount)
  {
    qWarning("min and max values for energy amount are equal!!!");
  }
}


GRadiatingAgent::~GRadiatingAgent()
{
}


void GRadiatingAgent::run()
{
  bool has_energy,has_form;
  has_energy = requestObject()->hasEnergy();
  has_form = requestObject()->hasForm();
  
  Q_ASSERT(has_energy);
  Q_ASSERT(has_form);

  if (!has_energy)
  {
    qWarning("can't radiate influence without energy!");
    return;
  } 
  
  while(!shutdown)
  {
    GEnergy* e = requestEnergy();
    if (e->amount() > MinEnergyAmount && e->amount() > 0)
    {
      double fraction = this->Fraction;
      
      if (e->amount() > MaxEnergyAmount && (e->amount() - MaxEnergyAmount)/e->amount() > fraction)
      {
        fraction = (e->amount() - MaxEnergyAmount)/e->amount();
//         qDebug(QString::number(fraction));
      }
      
      if ((e->amount() - MinEnergyAmount)/e->amount() < fraction)
      {
        fraction = (e->amount() - MinEnergyAmount)/e->amount();
//         qDebug(QString::number(fraction));
      }
      
      if (has_form)
      {
        if (ActiveElement)
        {
          GActiveElementInfluence influence( getElementID(), e->take(fraction), requestForm()->Position, requestForm()->Rotation, requestForm()->getRadiusMax(), requestObject()->getParent(), SenseDistance, SenseDistanceDetail );
          emit radiateInfluence( influence );
        }
        else
        {
          GPositionalInfluence influence( getElementID(), e->take(fraction), requestForm()->Position, requestForm()->Rotation, requestForm()->getRadiusMax(), requestObject()->getParent() );
          emit radiateInfluence( influence );
        }
      }
      else
      {
        GElementInfluence influence( getElementID(), e->take(fraction));
        emit radiateInfluence( influence );
      }
      emit energyChanged(*requestEnergy());
    }
    msleep(Msleep);
  }
}

void GRadiatingAgent::setActiveElementData(bool element_active, double sense_distance, double sense_distance_detail)
{
  if (!element_active)
  {
    this->ActiveElement = false;
  }
  else
  {
    this->SenseDistance = sense_distance;
    this->SenseDistanceDetail = sense_distance_detail;
  
    //set active AFTER sense distances were set as this routine could
    //be interrupted by the influence radiating thread!!
    this->ActiveElement = element_active;
  }
}

};
