/***************************************************************************
 *   Copyright (C) 2004 by Raphael Langerhorst                             *
 *   raphael-langerhorst@gmx.at                                            *
 *                                                                         *
 *   Permission is hereby granted, free of charge, to any person obtaining *
 *   a copy of this software and associated documentation files (the       *
 *   "Software"), to deal in the Software without restriction, including   *
 *   without limitation the rights to use, copy, modify, merge, publish,   *
 *   distribute, sublicense, and/or sell copies of the Software, and to    *
 *   permit persons to whom the Software is furnished to do so, subject to *
 *   the following conditions:                                             *
 *                                                                         *
 *   The above copyright notice and this permission notice shall be        *
 *   included in all copies or substantial portions of the Software.       *
 *                                                                         *
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,       *
 *   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF    *
 *   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.*
 *   IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR     *
 *   OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, *
 *   ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR *
 *   OTHER DEALINGS IN THE SOFTWARE.                                       *
 ***************************************************************************/

#include "GReparentAgent.h"
#include "GPositionalInfluence.h"
#include "GReparentInfluence.h"

#include <GObject.h>
#include <GElementInfluence.h>

#include <qvaluelist.h>

#include <exception>

using namespace GCS;

namespace GBE {

GReparentAgent::GReparentAgent(QObject* parent, const char* name)
 : GCS::GAgent(parent,name)
{
}


GReparentAgent::~GReparentAgent()
{
}

void GReparentAgent::performReparenting(const GElementID& new_parent, const GVector3& pos_add, const GVector3& rotate)
{
  GObject* obj = requestObject();
  Q_ASSERT(obj);
  const GElementID old_parent(obj->getParent());
  obj->reparent(new_parent);
  if (obj->hasForm())
  {
    GForm* f = requestForm();
    f->Position.turnAroundAxis(GVector3(1,0,0),rotate.x);
    f->Position.turnAroundAxis(GVector3(0,1,0),rotate.y);
    f->Position.turnAroundAxis(GVector3(0,0,1),rotate.z);
    f->Position.add(pos_add);
    f->Rotation.add(rotate);
  }
  else
  {
    qWarning("Element " + QString::number(getElementID().getID()) + " received a reparent influence but has no form!");
  }
  emit this->parentChanged(old_parent,new_parent);
}

void GReparentAgent::receiveInfluence(GCS::GElementInfluence& influence)
{
  try
  {
    GPositionalInfluence& pos_inf = dynamic_cast<GPositionalInfluence&>(influence);
    
    //we can only reparent either a child or this element if this element has a form
    if (requestObject()->hasForm())
    { 
      GForm* f = requestForm();
      const GElementID& source_id = pos_inf.source();
      
      //handle child reparenting
      if (this->requestObject()->getChildren().contains(source_id))
      {            
        //check if child is leaving me
        if (pos_inf.position().length() + pos_inf.radius() > requestForm()->getRadiusMax())
        {
          qDebug(QString::number(getElementID().getID()) + ": Reparenting child " + QString::number(source_id.getID()) + " from old parent (this): " + QString::number(requestObject()->getID().getID()) + " to own parent: " + QString::number(requestObject()->getParent().getID()));
          GVector3 rotation = f->Rotation;
          rotation.mul(-1);
          GReparentInfluence ri(getElementID(), GEnergy(0,0,0),
                                source_id,
                                requestObject()->getParent(),
                                f->Position,
                                rotation);
//           qDebug("Notify reparented element about reparenting.");
          emit sendInfluence(source_id,ri);
//           qDebug("Notify the new parent about the new child.");
          emit sendInfluence(requestObject()->getParent(),ri);
          
//           qDebug("Remove child from own element.");
          requestObject()->removeChild(ri.getReparentedElement());
        }
      }
      //handle reparenting this element (enters another element on same hierarchy level)
      else if (pos_inf.parent() == requestObject()->getParent() && pos_inf.source() != requestObject()->getParent() && pos_inf.source() != requestObject()->getID().getID())
      {
        GVector3 distance = pos_inf.position() - f->Position;
        if (pos_inf.radius() > distance.length() + f->getRadiusMax())
        {
          qDebug(QString::number(getElementID().getID()) + ": Reparenting this element " + QString::number(getElementID().getID()) + " from old parent: " + QString::number(requestObject()->getParent().getID()) + " to new parent: " + QString::number(pos_inf.source().getID()));
          GVector3 pos_add = pos_inf.position();
          pos_add.mul(-1);
          GVector3 rot_add = pos_inf.rotation();
          rot_add.mul(-1);
          GReparentInfluence ri(getElementID(), GEnergy(0,0,0),
                                getElementID(),
                                source_id,
                                pos_add,
                                rot_add);
                                
//           qDebug("Notify the old parent about the reparenting.");
          emit sendInfluence(requestObject()->getParent(),ri);
//           qDebug("Notify the new parent about the new child.");
          emit sendInfluence(source_id,ri);
          
//           qDebug("Perform form transformation and reparenting.");
          performReparenting(source_id,pos_add,rot_add);
          
//           qDebug("Parent is now: " + QString::number(requestObject()->getParent().getID()));
        }
      }
    }
  }
  catch(std::exception e)
  {
    //if it's not a positional influence, it might be a reparenting influence...
    try
    {
      GReparentInfluence& reparent_inf = dynamic_cast<GReparentInfluence&>(influence);
      
      GObject* obj = this->requestObject();
      if (obj)
      {
        //reparented element is this element - perform reparenting
        if (reparent_inf.getReparentedElement() == getElementID())
        {
          qDebug(QString::number(getElementID().getID()) + ": Reparenting this element: " + QString::number(getElementID().getID()));
          performReparenting(reparent_inf.getNewParent(),reparent_inf.getMoveDistance(),reparent_inf.getRotate()); 
        }
        //the new parent is this element - insert reparented element as child
        else if (reparent_inf.getNewParent() == getElementID())
        {
          qDebug(QString::number(getElementID().getID()) + ": Receiving new child which was reparented: " + QString::number(reparent_inf.getReparentedElement().getID()));
          obj->addChild(reparent_inf.getReparentedElement());
          //nothing to emit
        }
        //reparented element left this element (it was a child) - remove old child
        else if (requestObject()->getChildren().contains(reparent_inf.source()))
        {
          qDebug(QString::number(getElementID().getID()) + ": Removing old child which was reparented: " + QString::number(reparent_inf.getReparentedElement().getID()));
          requestObject()->removeChild(reparent_inf.source());
        }
        else
        {
          qWarning(QString::number(getElementID().getID()) + ": Received reparenting influence but don't know what to do with it!");
        }
      }
    }
    catch(std::exception e)
    {
    }
  }
}


};
