/***************************************************************************
 *   Copyright (C) 2004 by Raphael Langerhorst                             *
 *   raphael-langerhorst@gmx.at                                            *
 *                                                                         *
 *   Permission is hereby granted, free of charge, to any person obtaining *
 *   a copy of this software and associated documentation files (the       *
 *   "Software"), to deal in the Software without restriction, including   *
 *   without limitation the rights to use, copy, modify, merge, publish,   *
 *   distribute, sublicense, and/or sell copies of the Software, and to    *
 *   permit persons to whom the Software is furnished to do so, subject to *
 *   the following conditions:                                             *
 *                                                                         *
 *   The above copyright notice and this permission notice shall be        *
 *   included in all copies or substantial portions of the Software.       *
 *                                                                         *
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,       *
 *   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF    *
 *   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.*
 *   IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR     *
 *   OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, *
 *   ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR *
 *   OTHER DEALINGS IN THE SOFTWARE.                                       *
 ***************************************************************************/

#ifndef GBEGREPARENTAGENTH
#define GBEGREPARENTAGENTH

#include <GAgent.h>
#include <GElementID.h>
#include <GVector3.h>

#include <qobject.h>
#include <qmap.h>

namespace GBE {

/**
  \class GReparentAgent GReparentAgent.h
  \brief Handles element restructuring of child elements according to their positions.
  @author Raphael Langerhorst
  
  When receiving influences from child elements and these influences are
  GPositionalInfluence or compatible,
  the agent checks if the child is still in the context of itself - the
  child's current parent. If the child has left its current parent
  (according to position and size), a
  reparent influence is created and sent to the child and the new parent
  so they can change their child/parent settings.
  When the child has left this element it is reparented to this element's
  parent.
  
  If a reparent influence is received, the agent updates the parent/child
  information of the element and emits a reparent signal (NOT influence!)
  to let the GWE know about it.
  
  This agent also handles reparenting of itself into another child of the
  same parent. This happens when this element is completely enclosed in another
  element of the same parent. On such cases a reparenting influence is sent
  to the current parent and the new child so they can update their child/parent
  data.
  
  In all cases the reparented element has to perform form transformation
  because position and rotation is always relative to the current parent. The
  required transformation information is included in the reparenting influence
  which is sent by the element/agent that detects reparenting.
  
  Position and size information is stored in the form attribute of an element.
  GRadiatingAgent is taking care of influence radiation which is required
  for this agent to work as these influences include the position and size
  information.
  
  @see GReparentInfluence, GPositionalInfluence

*/

class GReparentAgent : public GCS::GAgent
{
  Q_OBJECT
  
  protected:
  
    /**
     * Do the actual reparenting. This is done by the reparented element.
     */
    void performReparenting(const GCS::GElementID& new_parent, const GCS::GVector3& pos_add, const GCS::GVector3& rotate);
    
  public:
  
    /**
     * Constructor.
     */
    GReparentAgent(QObject* parent = 0, const char* name = 0);

    /**
     * Virtual Destructor.
     */
    virtual ~GReparentAgent();
    
  public slots:
 
    /**
     * Analyses given influence and checks for reparenting if it's a
     * GPositionalInfluence or performs reparenting work if it is a
     * GReparentInfluence.
     */
    virtual void receiveInfluence(GCS::GElementInfluence& influence);

};

};

#endif
