/***************************************************************************
 *   Copyright (C) 2004 by Raphael Langerhorst                             *
 *   raphael-langerhorst@gmx.at                                            *
 *                                                                         *
 *   Permission is hereby granted, free of charge, to any person obtaining *
 *   a copy of this software and associated documentation files (the       *
 *   "Software"), to deal in the Software without restriction, including   *
 *   without limitation the rights to use, copy, modify, merge, publish,   *
 *   distribute, sublicense, and/or sell copies of the Software, and to    *
 *   permit persons to whom the Software is furnished to do so, subject to *
 *   the following conditions:                                             *
 *                                                                         *
 *   The above copyright notice and this permission notice shall be        *
 *   included in all copies or substantial portions of the Software.       *
 *                                                                         *
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,       *
 *   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF    *
 *   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.*
 *   IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR     *
 *   OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, *
 *   ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR *
 *   OTHER DEALINGS IN THE SOFTWARE.                                       *
 ***************************************************************************/


#include "OpenGLFrame.h"
#include <GElement.h>

#include <GObject.h>

#include <exception>

#include <qapplication.h>
#include <qptrstack.h>

using namespace std;

using namespace GCS;

namespace GCE
{


OpenGLFrame::OpenGLFrame(const GWEInterface* gwe, QWidget* parent, const char* name)
: QGLWidget(parent,name),
  Gwe(gwe)
{
  connect(qApp,SIGNAL(aboutToQuit()),this,SLOT(stopRendering()));
}


void OpenGLFrame::renderChildren(const GElement* element)
{

//   qDebug(QString("drawing child element: ") + QString::number(element->getElementID().getID()));

  QValueList<GElementID> children = element->getObject()->getChildren();
  
  for ( QValueListIterator<GElementID> it = children.begin(); it != children.end(); it++)
  {
    const GElement* element = Gwe->read((*it));
    
    Q_ASSERT(element);
    
    if (element)
    {
      
      const GObject* object = element->getObject();
      
      if (object->hasForm())
      {
      
        const GForm* f = object->getForm();
        
//         qDebug("element position: " + QString::number(f.Position.x) + ", " + QString::number(f.Position.y) + ", " + QString::number(f.Position.z));
      
        try
        {
          const GOpenGLForm* glf = dynamic_cast<const GOpenGLForm*>(f);

          glPushMatrix();

          glf->transform();
          renderChildren(element);
          if (glf->RGBA[3] > 0)
            glf->render();

          glPopMatrix();
        }
        catch (exception)
        {
          renderChildren(element);
        }
        
      }
      else
      {
        renderChildren(element);
      }
    }
  }
}

void OpenGLFrame::run()
{
  while(!stop_rendering)
  {
    this->update();
    msleep(this->sleep_time_ms);
  }
}

void OpenGLFrame::startRendering(int sleep_ms_between_updates_min)
{
  this->sleep_time_ms = sleep_ms_between_updates_min;
  this->stop_rendering = false;
  this->start();
}

void OpenGLFrame::stopRendering()
{
  this->stop_rendering = true;
  this->wait();
}

void OpenGLFrame::initializeGL()
{
  glClearColor(0,0,0,0);  //black background
  glShadeModel(GL_SMOOTH);
  glEnable(GL_DEPTH_TEST); //enable depth test
  glEnable(GL_BLEND); //enable blending - needed for alpha blending
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); //alpha blending
  glEnable(GL_LIGHTING); //globally enable lighting
  glEnable(GL_POLYGON_SMOOTH);
  
  //@todo: the light sources are ... a bit outdated, they were useful until 0.3.x versions. A new lighting model should be created.
  
  //create a light source
  glLightf(GL_LIGHT0,GL_CONSTANT_ATTENUATION,1);
  float v[] = {-1,5,-3,1};
  glLightfv(GL_LIGHT0,GL_POSITION,v);
  v[0] = v[1] = v[2] = 0.5;
  v[3] = 1;
  glLightfv(GL_LIGHT0,GL_AMBIENT,v);
  v[0] = 0.5;
  v[1] = 0.5;
  v[2] = 0.5;
  v[3] = 1;
  glLightfv(GL_LIGHT0,GL_DIFFUSE,v);
  glLightfv(GL_LIGHT0,GL_SPECULAR,v);
  glEnable(GL_LIGHT0);
  
  //another one:
  glLightf(GL_LIGHT1,GL_CONSTANT_ATTENUATION,1);
  v[0] = 0;
  v[1] = -5;
  v[2] = 3;
  v[3] = 1;
  glLightfv(GL_LIGHT1,GL_POSITION,v);
  v[0] = v[1] = v[2] = 0;
  v[3] = 0;
  glLightfv(GL_LIGHT1,GL_AMBIENT,v);
  v[0] = 0.5;
  v[1] = 0.5;
  v[2] = 0.5;
  v[3] = 1;
  glLightfv(GL_LIGHT1,GL_DIFFUSE,v);
  glLightfv(GL_LIGHT1,GL_SPECULAR,v);
  glEnable(GL_LIGHT1);
  
}

void OpenGLFrame::resizeGL(int w, int h)
{
  ViewProperties.width = w;
  ViewProperties.height = h;
  ViewProperties.update();
}

void OpenGLFrame::paintGL()
{
  QMutexLocker m(this);

  Q_ASSERT(Gwe);

  if (TopElements.isEmpty() || Gwe == NULL)
    return;
  
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
  glMatrixMode(GL_MODELVIEW);
  
  
  //draw top forms:
  for ( QValueListIterator<GElementID> it = TopElements.begin(); it != TopElements.end(); it++)
  {
    glLoadIdentity();
    
    //start with default values:
    GVector3 view_position(ViewProperties.ViewPosition);
    GVector3 view_target(ViewProperties.ViewTarget);
    GVector3 view_up(ViewProperties.ViewUp);
    
    //load element to which the camera is set:
    const GElement* camera_element = Gwe->read(ViewProperties.CameraElement);
    
    if (camera_element)
    {
      QPtrStack<const GElement> camera_element_hierarchy;
      camera_element_hierarchy.push(camera_element);
      
      //push parent elements onto stack until we either reach the current top element or there is no further parent
      while (camera_element && camera_element->getElementID() != camera_element->getObject()->getParent() && camera_element->getElementID() != (*it))
      {
        camera_element = Gwe->read(camera_element->getObject()->getParent());
        if (camera_element)
          camera_element_hierarchy.push(camera_element);
      }
      
      //now we have the whole element hierarchy that holds the camera element in one stack
      //so we can apply the camera transformations now to get final view positions and directions
      
      for (const GElement* el = camera_element_hierarchy.pop(); camera_element_hierarchy.isEmpty() == false; el = camera_element_hierarchy.pop())
      {
        const GObject* object = el->getObject();
        if (object->hasForm())
        {
          const GForm* f = object->getForm();
          //@todo we need to build translation and rotation matrix functionality into GMatrix44!!
          
          //first rotate
          view_position.turnAroundAxis(GVector3(1,0,0),f->Rotation.x);
          view_position.turnAroundAxis(GVector3(0,1,0),f->Rotation.y);
          view_position.turnAroundAxis(GVector3(0,0,1),f->Rotation.z);
          
          view_target.turnAroundAxis(GVector3(1,0,0),f->Rotation.x);
          view_target.turnAroundAxis(GVector3(0,1,0),f->Rotation.y);
          view_target.turnAroundAxis(GVector3(0,0,1),f->Rotation.z);
          
          view_up.turnAroundAxis(GVector3(1,0,0),f->Rotation.x);
          view_up.turnAroundAxis(GVector3(0,1,0),f->Rotation.y);
          view_up.turnAroundAxis(GVector3(0,0,1),f->Rotation.z);
          
          //then translate
          view_position.add(f->Position);
          view_target.add(f->Position);
        }
      }
      
    }
    else
    {
//      qWarning("No camera element set for the view properties or element not available!");
    }
    
    gluLookAt(view_position.x,view_position.y,view_position.z,
              view_target.x,view_target.y,view_target.z,
              view_up.x,view_up.y,view_up.z);
    
    const GElement* element = Gwe->read((*it));
    
    Q_ASSERT(element);
    
    if (element)
    {
      
      const GObject* object = element->getObject();
      
      if (object->hasForm())
      {
      
        const GForm* f = object->getForm();
        
//         qDebug("element position: " + QString::number(f.Position.x) + ", " + QString::number(f.Position.y) + ", " + QString::number(f.Position.z));

        try
        {
          const GOpenGLForm* glf = dynamic_cast<const GOpenGLForm*>(f);

          glPushMatrix();

          glf->transform();
          renderChildren(element);
          if (glf->RGBA[3] > 0)
            glf->render();

          glPopMatrix();
        }
        catch (exception)
        {
          //@todo currently we only transform when the form is a GOpenGLForm. We need to transform manually (per "software") when it's only a normal GForm!
          renderChildren(element);
        }
        
      }
      else
      {
        return renderChildren(element);
      }
    }
    
  }
  
  //at the end load an identitiy matrix
  glLoadIdentity();
}

/*
void OpenGLFrame::addTopForm(const GOpenGLForm* form)
{
  QMutexLocker m(this);
  TopForms.append(new FormNode(form));
}

void OpenGLFrame::removeTopForm(const GOpenGLForm* form)
{
  QMutexLocker m(this);
  
  //go through the list and remove the element
  FormNode* node = TopForms.first();
  while(node)
  {
    if (node->Form == form)
    {
      TopForms.remove();
      return;
    }
    node = TopForms.next();
  }
}

*/

void OpenGLFrame::addTopElement(const GElementID& element)
{
  QMutexLocker m(this);
  this->TopElements.append(element);
}

void OpenGLFrame::removeTopElement(const GElementID& element)
{
  QMutexLocker m(this);
  this->TopElements.remove(element);
}

}
