/***************************************************************************
 *   Copyright (C) 2003-2004 by Raphael Langerhorst                        *
 *   raphael-langerhorst@gmx.at                                            *
 *                                                                         *
 *   Permission is hereby granted, free of charge, to any person obtaining *
 *   a copy of this software and associated documentation files (the       *
 *   "Software"), to deal in the Software without restriction, including   *
 *   without limitation the rights to use, copy, modify, merge, publish,   *
 *   distribute, sublicense, and/or sell copies of the Software, and to    *
 *   permit persons to whom the Software is furnished to do so, subject to *
 *   the following conditions:                                             *
 *                                                                         *
 *   The above copyright notice and this permission notice shall be        *
 *   included in all copies or substantial portions of the Software.       *
 *                                                                         *
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,       *
 *   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF    *
 *   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.*
 *   IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR     *
 *   OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, *
 *   ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR *
 *   OTHER DEALINGS IN THE SOFTWARE.                                       *
 ***************************************************************************/


#include "GElement.h"

#include "GElementID.h"
#include "GObject.h"

#include "GElementInfluence.h"
#include "GAgent.h"

#include "GEnergy.h"

#include <qmutex.h>

#include <exception>

using namespace std;


namespace GCS
{

// ###################### PUBLIC ##################

GElement::GElement(GObject* object)
: Object(object)
{
  this->ParkTime = QDateTime::currentDateTime();
}

GElement::~GElement()
{
  //stop all running agents
  emit park();
    
  GAgent* agent = this->Agents.first();
  while(agent)
  {
    agent->wait();  //wait until thread is stopped before deleting
    agent->deleteLater();  //post a delete event (the event loop will delete it)
    Q_ASSERT(this->Agents.remove()); //print debug message if remove()==false
    agent = this->Agents.current();
  }

  Q_ASSERT(Object);
  if (Object)
    delete Object;
}

const GElementID& GElement::getElementID() const
{
  Q_ASSERT(Object);
  if (Object)
    return Object->getID();
  else
    throw exception();
}

bool GElement::isParked()
{
  //check all agents
  GAgent* a = Agents.first();
  while(a)
  {
    if (a->running())  //directly accesses the QThread method
      return false;
    a = Agents.next();
  }
  return true; //when no running agent was found
}

const QPtrList<GAgent>& GElement::getAgents() const
{
  return Agents;
}

const GObject* GElement::getObject() const
{
  Q_ASSERT(Object);
  if (Object)
    return Object;
  else
    throw new exception();
}

void GElement::parentChangedInternal(const GCS::GElementID& oldParent, const GCS::GElementID& newParent)
{
  emit this->parentChanged(this,oldParent,newParent);
}

// ################# PUBLIC SLOTS ####################

void GElement::receiveInfluence(GCS::GElementInfluence& influence)
{
  emit forwardInfluenceInternal(influence);
  
//this would be one possibility for the influence energy problem
//   if (Object)
//   {
//     if (Object->hasEnergy())
//     {
//       GEnergy e = Object->getEnergy();
//       QMutexLocker m(e);
//       e->put(influence.Energy);
//     }
//   }
}

void GElement::addAgent(GCS::GAgent* agent)
{
  agent->Object = this->Object;
  agent->Agents = &this->Agents;

  //connect signals from agent

  connect(agent,SIGNAL(
           sendInfluence(const GCS::GElementID&, GCS::GElementInfluence&)),
           this,SIGNAL(
           sendInfluence(const GCS::GElementID&, GCS::GElementInfluence&)));

  connect(agent,SIGNAL(radiateInfluence(GCS::GElementInfluence&)),
           this,SIGNAL(radiateInfluence(GCS::GElementInfluence&)));

  //connect slots from agent
  connect(this,SIGNAL(forwardInfluenceInternal(GCS::GElementInfluence&)),
         agent,SLOT(receiveInfluence(GCS::GElementInfluence&)));


  // control stuff (parking, executing)
   
  connect(this,SIGNAL(park()),
         agent,SLOT(beginPark()));

  // execute is used for "resuming" AND for starting the first time(!)
  connect(this,SIGNAL(execute(double)),
         agent,SLOT(threadStart(double)));
            
            
  // data change stuff
  
  connect(agent,SIGNAL(parentChanged(const GCS::GElementID&, const GCS::GElementID& )),
           this,SLOT(parentChangedInternal(const GCS::GElementID&, const GCS::GElementID& )));
  
  connect(agent,SIGNAL(agentChanged(const GCS::GAgent& )),
           this,SIGNAL(agentChanged(const GCS::GAgent& )));

  connect(agent,SIGNAL(energyChanged(const GCS::GEnergy& )),
           this,SIGNAL(energyChanged(const GCS::GEnergy& )));

  connect(agent,SIGNAL(formChanged(const GCS::GForm& )),
           this,SIGNAL(formChanged(const GCS::GForm& )));

  connect(agent,SIGNAL(childElementCreated(GCS::GElement* )),
           this,SIGNAL(childElementCreated(GCS::GElement* )));

  connect(agent,SIGNAL(childElementRemoved(const GCS::GElementID& )),
           this,SIGNAL(childElementRemoved(const GCS::GElementID& )));

  this->Agents.append(agent);
}

void GElement::removeAgent(GCS::GAgent* agent, bool del)
{
  //park the agent
  agent->beginPark();
  
  //wait until the agent is parked (the agent thread stops)
  agent->wait();

  agent->Object = NULL;
  agent->Agents = NULL;

  //disconnect agent  
  agent->disconnect(this);
  this->disconnect(agent);

  this->Agents.remove(agent);
  
  if (del)
    delete agent;
}

void GElement::parkElement()
{
  qDebug("Parking element " + QString::number(getElementID().getID()));
  emit park();
  this->ParkTime = QDateTime::currentDateTime();
}

void GElement::executeElement(double seconds_delta_t_offset)
{
  qDebug("Executing element " + QString::number(getElementID().getID()));
  QDateTime current = QDateTime::currentDateTime();
  double delta_t = (double)(this->ParkTime.secsTo(current)) + seconds_delta_t_offset;
  emit execute(delta_t);
}

}
