/***************************************************************************
 *   Copyright (C) 2004 by Raphael Langerhorst                             *
 *   raphael-langerhorst@gmx.at                                            *
 *                                                                         *
 *   Permission is hereby granted, free of charge, to any person obtaining *
 *   a copy of this software and associated documentation files (the       *
 *   "Software"), to deal in the Software without restriction, including   *
 *   without limitation the rights to use, copy, modify, merge, publish,   *
 *   distribute, sublicense, and/or sell copies of the Software, and to    *
 *   permit persons to whom the Software is furnished to do so, subject to *
 *   the following conditions:                                             *
 *                                                                         *
 *   The above copyright notice and this permission notice shall be        *
 *   included in all copies or substantial portions of the Software.       *
 *                                                                         *
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,       *
 *   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF    *
 *   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.*
 *   IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR     *
 *   OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, *
 *   ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR *
 *   OTHER DEALINGS IN THE SOFTWARE.                                       *
 ***************************************************************************/

#include <qapplication.h>
 
#include "CreateUniverseButton.h"

#include <GElement.h>
#include <GObject.h>
#include <GElementInfluence.h>

#include <GWEInterface.h>
#include <GWEInterfaceSimple.h>

#include <GMoveAgent.h>
#include <GInfluenceableAgent.h>
#include <GAttractAgent.h>
#include <GRadiatingAgent.h>

#include <InvestigationInterface.h>
#include <InvestigationInterfaceImp.h>
#include <GInvestigationAgent.h>
#include <GReparentAgent.h>
#include <GDynamicGeneratorAgent.h>
#include <GSolarsystemCategory.h>

#include <OpenGLFrame.h>
#include <Camera.h>

using namespace GCS;
using namespace GWE;
using namespace GBE;

using namespace GCE;


CreateUniverseButton::CreateUniverseButton(QSpinBox* spin_box_random_seed, QString text, QWidget *parent, const char *name)
 : QPushButton(text, parent, name),
   RandomSeed(spin_box_random_seed)
{
  connect(this,SIGNAL(pressed()),this,SLOT(createUniverse()));
}


CreateUniverseButton::~CreateUniverseButton()
{
}

QWidget* CreateUniverseButton::createUniverse()
{
  //create the G World Engine, it holds all element data and manages influences (routing between elements)    
  GWEInterface* gwe = new GWEInterfaceSimple();
  
  //create the 3D client interface, this is were everything is displayed
  OpenGLFrame* frame = new OpenGLFrame(gwe);
  
  //create a camera, the camera will work on the view properties of the frame
  Camera* cam = new Camera(frame);
    
  //to start with we want to create a galaxy:
  
  //a form with 100 000 units radius
  GOpenGLForm* myForm = new GOpenGLForm( GVector3(0,0,0), GVector3(0,0,0), 100000 );
  Q_ASSERT(myForm);
  
  //set alpha component to 0 (the borders should not be visible)
  myForm->RGBA[3] = 0;
  
  GElementID galaxy_id = GElementID::getNextFreeID();
  
  GObject* galaxy_object = new GObject(
                                new GEnergy(10,1000,3),
                                myForm,
                                GElementID(galaxy_id),  //the galaxy is it's own parent ;)
                                GElementID(galaxy_id),  //own ID is 1
                                GElementID(galaxy_id)   //connect to itself (connectionID is the same as own ID)
                              );
  
  //now create the solar system element
  GElement* galaxy = new GElement(galaxy_object);

  //add a generator agent to the element, it will take care of galaxy generation handling
  //including generation of solar systems,...
  QPtrList<GDynamicGeneratorCategory> categories;
  categories.append(new GSolarsystemCategory(0.5));
  galaxy->addAgent(new GDynamicGeneratorAgent(RandomSeed->value(),categories,QDateTime::currentDateTime(),7));
  
  //the galaxy should radiate some energy as well:
  galaxy->addAgent(new GRadiatingAgent(0.00002,10,1000,1000));
  
  //galaxy should handle reparenting
  galaxy->addAgent(new GReparentAgent());
                         
  //now we add the galaxy to the world engine, the world engine takes care
  //of influence routing, element storage, etc. (later also the network layer,...)
  //note that created children are automatically recognized by the GWE
  gwe->add(galaxy);
  
  
  //now we will also create an investigation element so we can analyse what's going on in
  //the universe...
  
  myForm = new GOpenGLForm(GVector3(0,0,-5),GVector3(0,0,0),0.01);
  Q_ASSERT(myForm);
//   myForm->RGBA[0] = myForm->RGBA[1] = myForm->RGBA[2] = 1;
  myForm->RGBA[3] = 0;
  
  cam->setForm(myForm);
  
  GElement* i_element = new GElement(
                           new GObject(      //create an object that holds most of the element data
                             new GEnergy(5,1,5),  //energy level, energy amount, sigma; big sigma is good ;-)
                             myForm,         //the form for the element, we created it above;
                             GElementID(galaxy_id),  //galaxy as parent
                             GElementID(GElementID::getNextFreeID()),  //own ID is 2
                             GElementID(galaxy_id)   //connect to the solar system (connectionID is the same as ID of the galaxy element)
                           )
                         );
                           
  //add some more agents to the element
  i_element->addAgent(new GInfluenceableAgent());
  
  //note that we currently use the camera to accelerate the speed of the form attribute.
  i_element->addAgent(new GMoveAgent(0.5));
  
  GRadiatingAgent* radiating_agent = new GRadiatingAgent(0.00002,0.001,10000,1000);
  radiating_agent->setActiveElementData(true,3000,500);
  i_element->addAgent(radiating_agent);
  
  //currently we do NOT reparent the investigation element
  //because of limitations in the camer to transform position
  //and rotation correctly - these are relative to the parent element.
  //please note that this can produces a corrupted parent/child hierarchy!
  //@todo improve client engine to be able to set camera on child elements
//   ii_element->addAgent(new GReparentAgent());
  
  //add the investigation element to the GWE
  gwe->add(i_element);
  
  //we need to tell the galaxy that the investigation element is it's child;
  galaxy_object->addChild(i_element->getElementID().getID());
    
  //set the frame as the main widget
  //@todo: we should create a sophisticated KDE interface for the next release
  
  
  //frame initialization
  frame->resizeGL(640,480);
  frame->resize(640,480);
  frame->initializeGL();
  
  frame->ViewProperties.NearClippingPlane = 0.1;
  frame->ViewProperties.FarClippingPlane = 10000;  //reduce this if you have slow rendering
  frame->ViewProperties.ViewPosition.z = 0.5;
  frame->ViewProperties.ViewTarget = GVector3(0,0,0);
  frame->ViewProperties.update();
  
  //tell the frame wich element it should render
  frame->addTopElement(galaxy->getElementID());
    
  //update the frame regularly:  
  frame->startRendering();
  //start camera handling:
  cam->startCamera();  

  //start element execution, this effectively starts the agent threads
  galaxy->executeElement();
  
  //run our investigation element
  i_element->executeElement();
  
  frame->show();
  
  return frame;
}
