/***************************************************************************
 *   Copyright (C) 2004 by Raphael Langerhorst                             *
 *   raphael-langerhorst@gmx.at                                            *
 *                                                                         *
 *   Permission is hereby granted, free of charge, to any person obtaining *
 *   a copy of this software and associated documentation files (the       *
 *   "Software"), to deal in the Software without restriction, including   *
 *   without limitation the rights to use, copy, modify, merge, publish,   *
 *   distribute, sublicense, and/or sell copies of the Software, and to    *
 *   permit persons to whom the Software is furnished to do so, subject to *
 *   the following conditions:                                             *
 *                                                                         *
 *   The above copyright notice and this permission notice shall be        *
 *   included in all copies or substantial portions of the Software.       *
 *                                                                         *
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,       *
 *   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF    *
 *   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.*
 *   IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR     *
 *   OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, *
 *   ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR *
 *   OTHER DEALINGS IN THE SOFTWARE.                                       *
 ***************************************************************************/

#include "GWEInterfaceSimple.h"

#include <GObject.h>
#include <GForm.h>
#include <GElement.h>
#include <GElementID.h>
#include <GVector3.h>
#include <GMatrix44.h>


using namespace GCS;
 
namespace GWE
{

GWEInterfaceSimple::GWEInterfaceSimple()
{
  this->ElementListMutex = new QMutex();
}

GWEInterfaceSimple::~GWEInterfaceSimple()
{
  delete this->ElementListMutex;
  //@todo proper parking and storing (into DB?) of all elements!!!!!!!!!
}

bool GWEInterfaceSimple::add(GElement* element)
{
  QMutexLocker lock(this->ElementListMutex);
  
  connect(element,SIGNAL(parentChanged(GCS::GElement*, const GCS::GElementID&, const GCS::GElementID& )),
             this,SLOT(elementReparented(GCS::GElement*, const GCS::GElementID&, const GCS::GElementID& )));

  connect(element,SIGNAL(childElementCreated(GCS::GElement* )),
             this,SLOT(newElementCreated(GCS::GElement* )));
  connect(element,SIGNAL(childElementRemoved(const GCS::GElementID& )),
             this,SLOT(elementRemoved(const GCS::GElementID& )));
  
  connect(element,SIGNAL(radiateInfluence(GCS::GElementInfluence& )),
             this,SLOT(radiateInfluence(GCS::GElementInfluence& )));
  connect(element,SIGNAL(sendInfluence(const GCS::GElementID&, GCS::GElementInfluence& )),
             this,SLOT(routeInfluence(const GCS::GElementID&, GCS::GElementInfluence& )));
  
  Elements.prepend(element);
  qDebug("Added element with ID " + QString::number(element->getElementID().getID()));
  return true;
}

GElement* GWEInterfaceSimple::getElement(const GElementID& id) const
{
  QMutexLocker lock(ElementListMutex);
  QPtrListIterator<GElement> it( Elements );
  GElement* element;
  while( (element = it.current()) != 0)
  {
    ++it;
    if (element->getElementID() == id)
      return element;
  }
  
  return NULL;
}

QPtrList<GElement> GWEInterfaceSimple::findInRange(GElement* source)
{
  QPtrList<GElement> list;
  
  Q_ASSERT(source);
  
  if (source==NULL)
  {
    qWarning("source is NULL!!!");
    return list;
  }
    
  const GObject* source_o = source->getObject();
  
  const GElementID& parentID = source->getObject()->getParent();
  
  GElement* parent = getElement(parentID);
  
  Q_ASSERT(parent);
        
  //now, getParent()->children() to get a list of elements with the
  //same parent as source
  
  QValueList<GElementID> children;
  
  if (parent)
  {
    //parent is affected as well
    list.append(parent);
    
    //if the element is it's own parent we do not need to go through
    //it's children here, they will all be added later anyway
    if (parentID != source->getElementID().getID())
    {
    
      children = parent->getObject()->getChildren();
      
      bool has_form = source_o->hasForm() ? true : false;
      
      if (has_form) //if the source element has a form, we need to check if they touch each other
      {
        const GForm* source_f = source_o->getForm();
        const GVector3& position = source_f->Position;
        const double range = source_f->getRadiusMax();
      
        for (QValueListIterator<GElementID> childrenID = children.begin(); childrenID != children.end(); ++childrenID)
        {
          GElement* el = getElement(*childrenID);
          Q_ASSERT(el);
          if (el)
          {
            const GObject* el_o= el->getObject();
            if (el_o->hasForm())
            {
              const GForm* f = el_o->getForm();
              if ((f->Position - position).length() < (range + f->getRadiusMax()))
                list.append(el);
            }
            else
            {
              //if something has no form it is everywhere
              list.append(el);
            }
          }
        }
      }
      else //if the source element has no form, it reaches everything anyway
      {
        for (QValueListIterator<GElementID> childrenID = children.begin(); childrenID != children.end(); ++childrenID)
        {
          GElement* el = getElement(*childrenID);
          Q_ASSERT(el);
          if (el)
          {
            list.append(el);
          }
        }
      }
    }
  }
  else
  {
    qDebug("element with ID " + QString::number(source->getElementID().getID()) + " has no parent!");
    //the element at least influences itself
    list.append(source);
  }
  
  //add own children - the element that radiates a certain influence DOES influence "itself" by this
  children = source->getObject()->getChildren();
  
  for (QValueListIterator<GElementID> childrenID = children.begin(); childrenID != children.end(); ++childrenID)
  {
    GElement* el = getElement(*childrenID);
    Q_ASSERT(el);
    if (el)
    {
      list.append(el);
    }
  }
  
  //this list contains:
  // * the parent
  // * all children of it's parent that touch the source's form (including itself)
  // * all own children
  return list;
}

const GElement* GWEInterfaceSimple::read(const GElementID& id) const
{
  return getElement(id);
}

GElement* GWEInterfaceSimple::open(const GElementID& id)
{
  GElement* element = getElement(id);
  if (element)
  {
    emit this->elementOpened(element);
  }
  
  return element;
}

bool GWEInterfaceSimple::close(GElement* element)
{
  emit this->elementClosed(element);
  return true;
}

bool GWEInterfaceSimple::postDelete(GElement* element)
{
  close(element);
  element->deleteLater();
  return Elements.remove(element);
}

void GWEInterfaceSimple::newElementCreated(GElement* element)
{
  Q_ASSERT(element);
  if (element)
  {
    add(element);
  }
  else
    qWarning("received new element, but element was NULL");
}

void GWEInterfaceSimple::elementRemoved(const GElementID& ID)
{
  postDelete(getElement(ID));
}

void GWEInterfaceSimple::elementReparented(GCS::GElement* element, const GCS::GElementID& oldParent, const GCS::GElementID& newParent)
{
  //
}

void GWEInterfaceSimple::radiateInfluence(GElementInfluence& influence)
{
  GElement* source_element = getElement(influence.source());
  
//   qDebug("got influence to radiate from " + QString::number(source_element->getElementID().getID()));
  
  QPtrList<GElement> influenced_elements = this->findInRange(source_element);
  
  //for now provide a very simple mechanism for influence distribution:
  // energy amount of the influence is evenly distributed among its receivers
  
  unsigned dest_count = influenced_elements.count();
  
  if (dest_count > 0)
  {
    double energy_level_original = influence.Energy.level();
    double energy_amount_original = influence.Energy.amount();
    double energy_sigma_original = influence.Energy.sigma();
    
    double energy_amount_single_dest = energy_amount_original/dest_count;
    
    double energy_amount_left = 0;
    
    GElement* e = influenced_elements.first();
    while(e)
    {
      influence.Energy.set(energy_level_original,energy_amount_single_dest,energy_sigma_original);
//       qDebug(" sending to " + QString::number(e->getElementID().getID()));
      e->receiveInfluence(influence);
      energy_amount_left += influence.Energy.amount();
      e = influenced_elements.next();
    }
    
    //set the energy amount to what has been left by the receivers
    //the sender should put this energy back to it's own energy
    influence.Energy.set(energy_level_original,energy_amount_left,energy_sigma_original);
  }
  else
  {
    qWarning("destination count for influence radiation is 0!");
  }
}

void GWEInterfaceSimple::routeInfluence(const GElementID& destination, GElementInfluence& influence)
{

//   qDebug("got influence to route from " + QString::number(influence.source().getID()));

  GElement* element = getElement(destination);
  Q_ASSERT(element);
  if (element)
  {
    element->receiveInfluence(influence);
  }
  else
  {
    element = getElement(influence.source());
    if (element==NULL)
    {
      Q_ASSERT(element);
      qDebug("got influence from %lu, but this element is not stored?!",influence.source().getID());
    }
    else
    {
//       qDebug(" sending to " + QString::number(element->getElementID().getID()));
      element->receiveInfluence(influence);
    }
  }
}

}
