/***************************************************************************
 *   Copyright (C) 2004 by Raphael Langerhorst                             *
 *   raphael-langerhorst@gmx.at                                            *
 *                                                                         *
 *   Permission is hereby granted, free of charge, to any person obtaining *
 *   a copy of this software and associated documentation files (the       *
 *   "Software"), to deal in the Software without restriction, including   *
 *   without limitation the rights to use, copy, modify, merge, publish,   *
 *   distribute, sublicense, and/or sell copies of the Software, and to    *
 *   permit persons to whom the Software is furnished to do so, subject to *
 *   the following conditions:                                             *
 *                                                                         *
 *   The above copyright notice and this permission notice shall be        *
 *   included in all copies or substantial portions of the Software.       *
 *                                                                         *
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,       *
 *   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF    *
 *   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.*
 *   IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR     *
 *   OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, *
 *   ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR *
 *   OTHER DEALINGS IN THE SOFTWARE.                                       *
 ***************************************************************************/


#ifndef GWEINTERFACESIMPLEH
#define GWEINTERFACESIMPLEH


#include "GWEInterface.h"

#include <GMatrix44.h>

#include <qptrlist.h>
#include <qmutex.h>


namespace GWE
{

/**
  \class GWEInterfaceSimple GWEInterfaceSimple.h
  \brief Provides local non-persistent GWE functionality
  @author Raphael Langerhorst
  
  GWEInterfaceSimple implements the GWEInterface interface,
  all functionality provided in this class works as stated
  in the documentation of the GWEInterface class;
  
  the only exception is that data can't be kept persistent -
  all elements are only held in application memory and don't
  survive a shutdown/restart cycle.
*/

class GWEInterfaceSimple : public GWEInterface
{
  Q_OBJECT
  private:
  
    /**
     * Represents the whole storage mechanism for this "simple"
     * GWE.
     * @todo use some hash table for this (GElementID as identifier)...
     */
    QPtrList<GCS::GElement> Elements;
    
    /**
     * Used for locking the elements list.
     * This is important for threaded access.
     */
    QMutex* ElementListMutex;
    
  protected:
    
    /**
     * Searches for the element with given ID and returns it.
     * If the element does not exist, NULL is returned.
     */
    GCS::GElement* getElement(const GCS::GElementID& id) const;
    
    /**
     * Returns a list of all elements that are in range of the source.
     * This means the source's parent, all children and all elements
     * that are in range of the source by checking the form attribute.
     * If the source element has no form it is expected to be present
     * in the whole parent, so all siblings are in range.
     * @note If the parent itself overlapts with another element then
     * only elements with the same parent as the source are listed, as
     * other elements don't have direct contact.
     */
    QPtrList<GCS::GElement> findInRange(GCS::GElement* source);
    
  
  public:
  
    /**
     * Constructor.
     */
    GWEInterfaceSimple();
  
    /**
     * Virtual destructor
     */
    virtual ~GWEInterfaceSimple();
  
    /**
     * @see GWEInterface::add()
     */
    virtual bool add(GCS::GElement*);
      
    /**
     * @see GWEInterface::read()
     */
    virtual const GCS::GElement* read(const GCS::GElementID&) const;
    
    /**
     * @see GWEInterface::open(), GWEInterface::close()
     */
    virtual GCS::GElement* open(const GCS::GElementID&);
    
    /**
     * @see GWEInterface::close(), GWEInterface::open()
     */
    virtual bool close(GCS::GElement*);
    
    /**
     * @see GWEInterface::postDelete()
     */
    virtual bool postDelete(GCS::GElement*);
    
  protected slots:
  
  // new element registration and element removal //
  
  // note that the source of these signals are active agents;
  
    /**
     * Called when a new element is being created
     * @note used internally (protected)
     */
    virtual void newElementCreated(GCS::GElement* element);
    
    /**
     * Called when a element should be removed
     * @note used internally (protected)
     */
    virtual void elementRemoved(const GCS::GElementID& ID);
    
    /**
     * Called when an element changes its parent.
     * @note used internally (protected)
     * @note Transformation is handled by element itself,
     * this slot just exists to allow information updating.
     * In a networked GWE this might cause responsibility
     * shifting between GWE server instances.
     */
    virtual void elementReparented(GCS::GElement* element, const GCS::GElementID& oldParent, const GCS::GElementID& newParent);
    
  public slots:
        
  // INFLUENCE MANAGEMENT //
  
    /**
     * @note To find all affected elements
     * GWEInterfaceSimple::findInRange is used.
     *
     * @see GWEInterface::radiateInfluence(), findInRange()
     */
    virtual void radiateInfluence(GCS::GElementInfluence& influence);
    
    /**
     * @see GWEInterface::routeInfluence()
     */
    virtual void routeInfluence(const GCS::GElementID& destination, GCS::GElementInfluence& influence);
};

}

#endif
